#include <math.h>

#include <glog/logging.h>

#include "actuator.h"
#include "simulator.h"
#include "util.h"

class Actuator1 : public Actuator {
public:
    Actuator1() {
        nextHohmann = 2;
        arrival = -1;
        goal = 0;
    }

    virtual int arrivalTime() {
        return arrival < 0 ? INT_MAX : arrival + baseTurn;
    }

    virtual void setGoal(double goal) {
        LOG(INFO) << "Setting goal=" << goal;
        this->goal = goal;
    }

    bool isOK(Simulator& sim) {
        int ok = 0;
        int best = -1;
        for (int t = 0; t < 10000; t++) {
            if (fabs(goal - dist(sim.x, sim.y)) < 800) {
                ok++;
                if (ok > 903) {
                    sim.restore();
                    return true;
                }
            } else {
                ok = 0;
            }
            sim.step();
        }
        sim.restore();
        return false;
    }

    int calcBest(Simulator& sim, int cnt, double scale,
                 double* dvx, double* dvy) {
        int best = -1;
        *dvx = 0;
        *dvy = 0;
        for (int i = 0; i <= cnt; i++) {
            double ndvx = (i / (double)cnt - 0.5) * scale;
            for (int j = 0; j <= cnt; j++) {
                double ndvy = (i / (double)cnt - 0.5) * scale;
                int ok = 0;
                sim.step(ndvx, ndvy);
                for (int t = 0; t < 5000; t++) {
                    if (fabs(goal - dist(sim.x, sim.y)) < 800) {
                        ok++;
                        if (best < ok ||
                            (best == ok && dist(ndvx, ndvy) > dist(*dvx, *dvy))) {
                            best = ok;
                            *dvx = ndvx;
                            *dvy = ndvy;
                        }
                    } else {
                        ok = 0;
                    }
                    sim.step();
                }

                sim.restore();
            }
        }
        return best;
    }

    virtual void step_() {
        dvx = dvy = 0;
        if (turn == 1 && goal == 0) {
            setGoal(out[4]);
        }
        else if (turn == nextHohmann) {
            LOG(INFO) << "Let's plan the trip! turn=" << turn;

            double x = out[2];
            double y = out[3];

            double r0 = dist(x, y);
            double rt = goal;

#if 0
            if (out[1] == 10000.0) {
                rt = r0 * 2;
            }
#endif

            double t = M_PI * sqrt(cube(r0 + rt) / 8 / MU);
            double dv0 = sqrt(MU / r0) * (sqrt(2 * rt / (r0 + rt)) - 1);
            double dvt = sqrt(MU / rt) * (1 - sqrt(2 * r0 / (r0 + rt)));

            double dv0x = sdiv(-dv0 * y, r0);
            double dv0y = sdiv(dv0 * x, r0);
            double dvtx = sdiv(dvt * y, r0);
            double dvty = sdiv(-dvt * x, r0);

            clockwise = isClockwise(px, py, x, y);
            if (clockwise) {
                dv0x = -dv0x;
                dvtx = -dvtx;
            }

            double resource = dist(dv0x, dv0y);

            addPlan(turn, 2, dv0x);
            addPlan(turn, 3, dv0y);
            addPlan(turn+1, 2, 0);
            addPlan(turn+1, 3, 0);

            t += turn-1;
            int it = (int)t;
            double dt = t - it;
            addPlan(it, 2, dvtx * (1 - dt));
            addPlan(it, 3, dvty * (1 - dt));
            addPlan(it + 1, 2, dvtx * dt);
            addPlan(it + 1, 3, dvty * dt);

            resource += dist(dvtx * (1 -dt), dvty * (1 - dt));
            resource += dist(dvtx * dt, dvty * dt);

            vector<double> planX;
            vector<double> planY;

            Simulator sim(px, py, x, y);
            sim.step(dv0x, dv0y);
            LOG(INFO) << "Current distance: " << dist(sim.x, sim.y) - goal;
            for (int i = 1; i < it; i++) sim.step();
            LOG(INFO) << "Current distance: " << dist(sim.x, sim.y) - goal;
            sim.step(dvtx * (1 - dt), dvty * (1 - dt));
            sim.step(dvtx * dt, dvty * dt);
            sim.save();

            LOG(INFO) << "Current distance: " << dist(sim.x, sim.y) - goal;

            if (!isOK(sim)) {
                for (int j = 0; j < 200; j++) {
                    double dvx, dvy;
                    int best = calcBest(sim, 10, 1, &dvx, &dvy);
                    LOG(INFO) << "best=" << best
                              << " dvx=" << dvx << " dvy=" << dvy;
                    planX.push_back(dvx);
                    planY.push_back(dvy);
                    resource += dist(dvx, dvy);
                    sim.step(dvx, dvy);
                    sim.save();
                }
            }

            if (isOK(sim)) {
                double rest = out[1] - resource - 0.00001;
                LOG(INFO) << "Rest energy: " << rest;
                for (int i = 0; i < 1000; i++) {
                    Simulator nsim(sim);
                    double stx = nsim.x;
                    double sty = nsim.y;
                    nsim.step();
                    double dx = nsim.x - stx;
                    double dy = nsim.y - sty;
                    double dd = dist(dx, dy);
                    nsim.restore();
                    double bx = rest / 2 * dx / dd;
                    double by = rest / 2 * dy / dd;
                    nsim.step(bx, by);
                    nsim.step(-bx, -by);
                    nsim.save();
                    if (isOK(nsim)) {
                        planX.push_back(bx);
                        planY.push_back(by);
                        planX.push_back(-bx);
                        planY.push_back(-by);

                        break;
                    }
                    rest -= 100;
                    //fprintf(stderr, "retry\n");
                    if (rest <= 0) break;
                }
            }

            CHECK_EQ(planX.size(), planY.size());

            for (size_t j = 0; j < planX.size(); j++) {
                addPlan(it + 2 + j, 2, planX[j]);
                addPlan(it + 2 + j, 3, planY[j]);
            }
            addPlan(it + 2 + planX.size(), 2, 0);
            addPlan(it + 2 + planY.size(), 3, 0);

            arrival = it + 5;
            nextHohmann = it + 2000;
        }
        else {
#if 0
            double x = out[2];
            double y = out[3];
            double d = dist(x, y);
            double dr = fabs(goal - d);
            if (dr < 100) {
                burn();
            }
            else if (dr < 10000 && plans.empty()) {
                adjust();
            }
#endif
        }

        px = out[2];
        py = out[3];
        pdvx = dvx;
        pdvy = dvy;
    }

    void burn() {
    }

    void adjust() {
        double x = out[2];
        double y = out[3];
        double d = dist(x, y);

        Simulator sim(px, py, x, y, pdvx, pdvy);
        double v = sim.velocity();
        double real = sim.angle();

        sim.step();
        double r = atan2(sim.y, sim.x);

        double ideal = addAngle(r, clockwise ? M_PI * 0.5 : -M_PI * 0.5);
        LOG(INFO) << "ideal=" << ideal << " real=" << real << " r=" << r;
        int adjType = 0;
        if (real > ideal) {
            if ((clockwise && d < goal) || (!clockwise && d > goal)) {
                adjType = -1;
            }
        }
        else if (real < ideal) {
            if ((clockwise && d > goal) || (!clockwise && d < goal)) {
                adjType = 1;
            }
        }

        if (adjType) {
            if (fabs(d - goal) >= 100) {
                ideal += adjType * M_PI * 0.00001;
            }

            double ivx = cos(ideal) * v;
            double ivy = sin(ideal) * v;
            LOG(INFO) << "vx=" << sim.vx << " vy=" << sim.vy
                      << " ivx=" << ivx << " ivy=" << ivy;
            dvx = ivx - sim.vx;
            dvy = ivy - sim.vy;

            LOG(INFO) << "Adjust dvx=" << dvx << " dvy=" << dvy;
            addPlan(turn, 2, -dvx * 0.5);
            addPlan(turn, 3, -dvy * 0.5);
        }
    }

private:
    double px;
    double py;
    double dvx;
    double dvy;
    double pdvx;
    double pdvy;
    double goal;
    bool clockwise;
    int nextHohmann;
    int arrival;
};

Actuator* createActuator1() {
    return new Actuator1();
}
