#include <math.h>

#include <iomanip>
#include <vector>

#include <glog/logging.h>

#include "actuator.h"
#include "simulator.h"
#include "util.h"

using namespace std;

class Actuator2 : public Actuator {
public:
    Actuator2() {
        nextHohmann = 2;
    }

    virtual void step_() {
        double x = out[2];
        double y = out[3];
        double ox = x - out[4];
        double oy = y - out[5];

        if (turn == nextHohmann) {
            LOG(INFO) << "Let's plan the trip";

            static const int N = 100000;

            Simulator sim(px, py, x, y);
            //LOG(INFO) << POS(sim.vx, sim.vy);
            vector<double> xa;
            vector<double> ya;
            sim.future(N, &xa, &ya);

            Simulator osim(pox, poy, ox, oy);
            vector<double> oxa;
            vector<double> oya;
            vector<double> ovxa;
            vector<double> ovya;
            osim.future(N + 5000, &oxa, &oya, &ovxa, &ovya);
            //LOG(INFO) << POS(osim.vx, osim.vy);

            double r0 = dist(x, y);
            double rt = dist(ox, oy);
            double rr = rt / r0;

            double t = M_PI * sqrt(cube(r0 + rt) / 8 / MU);
            double dv0 = sqrt(MU / r0) * (sqrt(2 * rt / (r0 + rt)) - 1);
            double dvt = sqrt(MU / rt) * (1 - sqrt(2 * r0 / (r0 + rt)));

            int it = (int)t;
            double dt = t - it;
            int start = -1;
            for (int i = 0; i < N - it - 1; i++) {
                double nx = -xa[i] * rr;
                double ny = -ya[i] * rr;
                double nox = oxa[i+it] * (1 - dt) + oxa[i+it+1] * dt;
                double noy = oya[i+it] * (1 - dt) + oya[i+it+1] * dt;

                double dx = nx - nox;
                double dy = ny - noy;
                double d = dist(dx, dy);
                //fprintf(stderr, "%f\n", d);
                if (d < 1000) {
                    start = i;
                    break;
                }
            }

            //CHECK_NE(start, -1);
            if (start == -1) return;

            double dv0x = sdiv(-dv0 * ya[start], r0);
            double dv0y = sdiv(dv0 * xa[start], r0);
            double dvtx = sdiv(dvt * ya[start], r0);
            double dvty = sdiv(-dvt * xa[start], r0);

            clockwise = isClockwise(px, py, x, y);
            if (clockwise) {
                dv0x = -dv0x;
                dvtx = -dvtx;
            }

            if (!decidePlan(xa, ya, oxa, oya, ovxa, ovya,
                            start, t, dv0x, dv0y, dvtx, dvty)) {
                setPlans(start, t, dv0x, dv0y, dvtx, dvty);
            }
        }

        px = x;
        py = y;
        pox = ox;
        poy = oy;
    }

    void setPlans(int start, double t,
                  double dv0x, double dv0y, double dvtx, double dvty) {
        addPlan(turn+start, 2, dv0x);
        addPlan(turn+start, 3, dv0y);
        addPlan(turn+start+1, 2, 0);
        addPlan(turn+start+1, 3, 0);

        t += turn + start;
        int it = (int)t;
        double dt = t - it;
        addPlan(it, 2, dvtx * (1 - dt));
        addPlan(it, 3, dvty * (1 - dt));
        addPlan(it + 1, 2, dvtx * dt);
        addPlan(it + 1, 3, dvty * dt);
        addPlan(it + 2, 2, 0);
        addPlan(it + 2, 3, 0);

        nextHohmann = it + 2000;

        LOG(INFO) << "We start at " << (int)(turn+start)
                  << " with " << POS(dv0x, dv0y)
                  << " and will arrive at " << (int)(t)
                  << " with " << POS(dvtx, dvty)
                  << " next trip will be " << nextHohmann;
        ;
    }

    bool willSucceed(double x0, double y0, double x1, double y1,
                     const vector<double>& oxa, const vector<double>& oya,
                     int start, double t,
                     double dv0x, double dv0y, double dvtx, double dvty,
                     double* best) {
        Simulator sim(x0, y0, x1, y1);
        //LOG(INFO) << POS(sim.x, sim.y);
        sim.step(dv0x, dv0y);
        for (int i = 1; i < t-1; i++) {
            //LOG(INFO) << "turn=" << turn+start+i << POS(sim.x, sim.y);
            sim.step();
        }
        double dt = t - (int)t;
        //LOG(INFO) << "turn=" << turn+start+t << POS(sim.x, sim.y);
        //LOG(INFO) << POS(dvtx * (1 - dt), dvty * (1 - dt));
        sim.step(dvtx * (1 - dt), dvty * (1 - dt));
        //LOG(INFO) << "turn=" << turn+start+t+1 << POS(sim.x, sim.y);
        //LOG(INFO) << POS(dvtx * dt, dvty * dt);
        sim.step(dvtx * dt, dvty * dt);
        int ok = 0;
        int ot = start + t + 2;
        *best = 1e300;
        for (int i = 0; i < 4000; i++) {
/*
            LOG(INFO) << "turn=" << turn+start+t+i+2
                      << POS(sim.x, sim.y) << " vs " << POS(oxa[ot+i], oya[ot+i]);
*/

            double d = dist(sim.x - oxa[ot + i], sim.y - oya[ot + i]);
            if (*best > d) {
                *best = d;
            }
            if (d < 1000) {
                ok++;
                //fprintf(stderr, "%d %f\n", ok, d);
                if (ok > 900) return true;
            }
            else {
                ok = 0;
            }
            sim.step();
        }
        return false;
    }

    bool decidePlan(const vector<double>& xa, const vector<double>& ya,
                    const vector<double>& oxa, const vector<double>& oya,
                    const vector<double>& ovxa, const vector<double>& ovya,
                    int start, double t,
                    double dv0x, double dv0y, double dvtx, double dvty) {
        double px, py;
        if (start > 0) {
            px = xa[start-1];
            py = ya[start-1];
        }
        else {
            px = this->px;
            py = this->py;
        }

        double best;
        if (willSucceed(px, py, xa[start], ya[start], oxa, oya,
                        start, t, dv0x, dv0y, dvtx, dvty, &best)) {
            return false;
        }

        double bestX = 1e300;
        double bestDv0x = 0;
        for (int i = 0; i <= 100; i++) {
            double ndv0x = dv0x - 0.5 + i * 0.01;
            willSucceed(px, py, xa[start], ya[start], oxa, oya,
                        start, t, ndv0x, dv0y, 0, 0, &best);
            if (bestX > best) {
                bestX = best;
                bestDv0x = ndv0x;
            }
        }

        double bestY = 1e300;
        double bestDv0y = 0;
        for (int i = 0; i <= 100; i++) {
            double ndv0y = dv0y - 0.5 + i * 0.01;
            willSucceed(px, py, xa[start], ya[start], oxa, oya,
                        start, t, dv0x, ndv0y, 0, 0, &best);
            if (bestY > best) {
                bestY = best;
                bestDv0y = ndv0y;
            }
        }

        willSucceed(px, py, xa[start], ya[start], oxa, oya,
                    start, t, bestDv0x, bestDv0y, 0, 0, &best);
        LOG(INFO) << "best=" << best << " bestX="
                  << bestX << " bestY=" << bestY;
        if (best > bestX && bestY > bestX) {
            bestDv0y = dv0y;
            best = bestX;
        }
        else if (best > bestY && bestX > bestY) {
            bestDv0x = dv0x;
            best = bestY;
        }

        addPlan(turn+start, 2, bestDv0x);
        addPlan(turn+start, 3, bestDv0y);
        addPlan(turn+start+1, 2, 0);
        addPlan(turn+start+1, 3, 0);

        Simulator sim(px, py, xa[start], ya[start]);
        sim.step(bestDv0x, bestDv0y);
        int i;
        for (i = 1;; i++) {
            CHECK_LT(start+i, oxa.size());

            double d = dist(sim.x - oxa[start+i], sim.y - oya[start+i]);
            if (d == best) break;

            sim.save();
            sim.step();
        }

        sim.restore();
        double dvx = ovxa[start+i] - sim.vx;
        double dvy = ovya[start+i] - sim.vy;

        addPlan(turn+start+i-1, 2, -dvx);
        addPlan(turn+start+i-1, 3, -dvy);

#if 0
        i++;
        sim.step(-dvx/2, -dvy/2);

        dvx = ovxa[start+i] - sim.vx;
        dvy = ovya[start+i] - sim.vy;

        addPlan(turn+start+i, 2, -dvx);
        addPlan(turn+start+i, 3, -dvy);
#endif

        addPlan(turn+start+i, 2, 0);
        addPlan(turn+start+i, 3, 0);

        nextHohmann = turn+start+i+5000;

        return true;
    }

    double px, py;
    double pox, poy;
    int nextHohmann;
    bool clockwise;
};

Actuator* createActuator2() {
    return new Actuator2();
}
