#include <math.h>

#include <iomanip>
#include <vector>
#include <map>

#include <glog/logging.h>

#include "actuator.h"
#include "simulator.h"
#include "util.h"

using namespace std;

class Actuator3 : public Actuator {
    typedef map<double, pair<double, double> > Ellipse;

public:
    Actuator3() {
        nextHohmann = 2;
    }

    virtual void step_() {
        double x = out[2];
        double y = out[3];
        double ox = x - out[4];
        double oy = y - out[5];

        if (turn == arrival) {
            LOG(INFO) << "The expected arrival time: "
                      << POS(x, y) << " vs " << POS(ox, oy)
                      << " " << dist(x-ox, y-oy);
        }

        if (turn == nextHohmann) {
            LOG(INFO) << "Let's plan the trip turn=" << turn;

            static const int N = 1000000;

            Simulator sim(px, py, x, y);
            //LOG(INFO) << POS(sim.vx, sim.vy);
            vector<double> xa;
            vector<double> ya;
            sim.future(N, &xa, &ya);

            Simulator osim(pox, poy, ox, oy);
            vector<double> oxa;
            vector<double> oya;
            vector<double> ovxa;
            vector<double> ovya;
            osim.future(N + 5000, &oxa, &oya, &ovxa, &ovya);
            //LOG(INFO) << POS(osim.vx, osim.vy);

            Ellipse ellipse;
            for (size_t i = 0; i < oxa.size(); i++) {
                double ox = oxa[i];
                double oy = oya[i];
                double r = atan2(oy, ox);
                ellipse.insert(make_pair(r, make_pair(ox, oy)));
            }

            int start = -1;
            double d, dd, t;
            // Find good position
            for (int i = 0; i < N; i++) {
                double x = xa[i];
                double y = ya[i];
                double r = atan2(y, x);
                double dr = r + M_PI;
                if (dr > M_PI) dr -= M_PI * 2;

                Ellipse::const_iterator found = ellipse.lower_bound(dr);
                Ellipse::const_iterator found2(found);
                ++found2;
                if (found2 == ellipse.end()) found2 = ellipse.begin();

                double dx = found->second.first;
                double dy = found->second.second;

                dd = sqrt(dx * dx + dy * dy);
                d = sqrt(x * x + y * y);

                double ddd = d + dd;

                t = M_PI * sqrt(ddd*ddd*ddd/8/MU);
                size_t tt = (size_t)(i + t);
                if (tt+1 >= oxa.size()) {
                    continue;
                }

                double dt = i + t - tt;
                double oxe = oxa[tt] * (1 - dt) + oxa[tt+1] * dt;
                double oye = oya[tt] * (1 - dt) + oya[tt+1] * dt;
                double expected = dist(dx - oxe, dy - oye);

                if (expected < 1000) {
                    LOG(INFO) << "Found";
                    start = i;
                    break;
                }
            }

            //CHECK_NE(start, -1);
            if (start == -1) {
                LOG(INFO) << "Not found :(";
                nextHohmann = turn + N / 2;
                //nextHohmann = turn + 1000;
                return;
            }

            double dv0 = sqrt(G*M/d) * (sqrt(2*dd/(d+dd)) - 1);
            double dvt = sqrt(G*M/dd) * (1 - sqrt(2*d/(d+dd)));

            double r0 = dist(x, y);
            double dv0x = sdiv(-dv0 * ya[start], r0);
            double dv0y = sdiv(dv0 * xa[start], r0);
            double dvtx = sdiv(dvt * ya[start], r0);
            double dvty = sdiv(-dvt * xa[start], r0);

            clockwise = isClockwise(px, py, x, y);
            if (clockwise) {
                dv0x = -dv0x;
                dvtx = -dvtx;
            }

            if (!decidePlan(xa, ya, oxa, oya, ovxa, ovya,
                            start, t, dv0x, dv0y, dvtx, dvty)) {
                setPlans(start, t, dv0x, dv0y, dvtx, dvty);
            }
        }

        px = x;
        py = y;
        pox = ox;
        poy = oy;
    }

    void setPlans(int start, double t,
                  double dv0x, double dv0y, double dvtx, double dvty) {
        addPlan(turn+start, 2, dv0x);
        addPlan(turn+start, 3, dv0y);
        addPlan(turn+start+1, 2, 0);
        addPlan(turn+start+1, 3, 0);

        t += turn + start;
        int it = (int)t;
        double dt = t - it;
        addPlan(it, 2, dvtx * (1 - dt));
        addPlan(it, 3, dvty * (1 - dt));
        addPlan(it + 1, 2, dvtx * dt);
        addPlan(it + 1, 3, dvty * dt);
        addPlan(it + 2, 2, 0);
        addPlan(it + 2, 3, 0);

        nextHohmann = it + 2000;

        LOG(INFO) << "We start at " << (int)(turn+start)
                  << " with " << POS(dv0x, dv0y)
                  << " and will arrive at " << (int)(t)
                  << " with " << POS(dvtx, dvty)
                  << " next trip will be " << nextHohmann;
        ;
    }

    bool willSucceed(double x0, double y0, double x1, double y1,
                     const vector<double>& oxa, const vector<double>& oya,
                     int start, double t,
                     double dv0x, double dv0y, double dvtx, double dvty,
                     double* best) {
        Simulator sim(x0, y0, x1, y1);
        //LOG(INFO) << POS(sim.x, sim.y);
        sim.step(dv0x, dv0y);
        for (int i = 1; i < t-1; i++) {
            //LOG(INFO) << "turn=" << turn+start+i << POS(sim.x, sim.y);
            sim.step();
        }
        double dt = t - (int)t;
        //LOG(INFO) << "turn=" << turn+start+t << POS(sim.x, sim.y);
        //LOG(INFO) << POS(dvtx * (1 - dt), dvty * (1 - dt));
        sim.step(dvtx * (1 - dt), dvty * (1 - dt));
        //LOG(INFO) << "turn=" << turn+start+t+1 << POS(sim.x, sim.y);
        //LOG(INFO) << POS(dvtx * dt, dvty * dt);
        sim.step(dvtx * dt, dvty * dt);
        int ok = 0;
        int ot = start + t + 2;
        *best = 1e300;
        for (int i = 0; i < 4000; i++) {
/*
            LOG(INFO) << "turn=" << turn+start+t+i+2
                      << POS(sim.x, sim.y) << " vs " << POS(oxa[ot+i], oya[ot+i]);
*/

            double d = dist(sim.x - oxa[ot + i], sim.y - oya[ot + i]);
            if (*best > d) {
                *best = d;
            }
            if (d < 1000) {
                ok++;
                //fprintf(stderr, "%d %f\n", ok, d);
                if (ok > 900) return true;
            }
            else {
                ok = 0;
            }
            sim.step();
        }
        return false;
    }

    double adjustDv(const vector<double>& xa, const vector<double>& ya,
                    const vector<double>& oxa, const vector<double>& oya,
                    double px, double py,
                    int start, double t,
                    double dv0x, double dv0y,
                    int cnt, double scale,
                    double* bestDv0xo, double* bestDv0yo) {
        double best;
        double bestX = 1e300;
        double bestDv0x = 0;
        for (int i = 0; i <= 100; i++) {
            double ndv0x = dv0x + (i / (double)cnt - 0.5) * scale;
            willSucceed(px, py, xa[start], ya[start], oxa, oya,
                        start, t, ndv0x, dv0y, 0, 0, &best);
            if (bestX > best) {
                bestX = best;
                bestDv0x = ndv0x;
            }
        }
        LOG(INFO) << "bestX=" << bestX << " bestDv0x=" << bestDv0x;

        double bestY = 1e300;
        double bestDv0y = 0;
        for (int i = 0; i <= 100; i++) {
            double ndv0y = dv0y + (i / (double)cnt - 0.5) * scale;
            willSucceed(px, py, xa[start], ya[start], oxa, oya,
                        start, t, dv0x, ndv0y, 0, 0, &best);
            if (bestY > best) {
                bestY = best;
                bestDv0y = ndv0y;
            }
        }
        LOG(INFO) << "bestY=" << bestX << " bestDv0y=" << bestDv0y;

        willSucceed(px, py, xa[start], ya[start], oxa, oya,
                    start, t, bestDv0x, bestDv0y, 0, 0, &best);
        LOG(INFO) << "best=" << best << " bestX="
                  << bestX << " bestY=" << bestY;
        if (best > bestX && bestY > bestX) {
            bestDv0y = dv0y;
            best = bestX;
        }
        else if (best > bestY && bestX > bestY) {
            bestDv0x = dv0x;
            best = bestY;
        }

        *bestDv0xo = bestDv0x;
        *bestDv0yo = bestDv0y;
        return best;
    }

    bool willSucceed2(Simulator& sim,
                      const vector<double>& oxa, const vector<double>& oya,
                      int start, double dt,
                      double dvtx, double dvty,
                      double* best) {
        //LOG(INFO) << "Distance: " << dist(sim.x-oxa[start], sim.y-oya[start]) << " " << ot;
        sim.step(dvtx * (1 - dt), dvty * (1 - dt));
        sim.step(dvtx * dt, dvty * dt);
        int ot = start + 1;
        int ok = 0;
        *best = 0;
        for (int i = 0; i < 4000; i++) {
/*
            LOG(INFO) << "turn=" << turn+start+t+i+2
                      << POS(sim.x, sim.y) << " vs " << POS(oxa[ot+i], oya[ot+i]);
*/

            double d = dist(sim.x - oxa[ot + i], sim.y - oya[ot + i]);
            //LOG(INFO) << "D " << d;
/*
            if (*best > d) {
                *best = d;
            }
*/
            if (d < 1000) {
                ok++;
                if (*best < ok) {
                    *best = ok;
                }
                //fprintf(stderr, "%d %f\n", ok, d);
                if (ok > 2000) {
                    sim.restore();
                    //LOG(INFO) << "OK!";
                    return true;
                }
            }
            else {
                ok = 0;
            }
            sim.step();
        }
        sim.restore();
        return false;
    }

    double adjustDv2(Simulator& sim,
                     const vector<double>& xa, const vector<double>& ya,
                     const vector<double>& oxa, const vector<double>& oya,
                     double px, double py,
                     int start, double t,
                     double dv0x, double dv0y,
                     int cnt, double scale,
                     double* bestDv0xo, double* bestDv0yo, double* dto) {
        double best;
        double bestX = 0;
        double bestDt = 0;
#if 1
        for (int i = 0; i <= cnt; i++) {
            double dt = i / (double)cnt;
            willSucceed2(sim, oxa, oya,
                         start, dt, dv0x, dv0y, &best);
            if (bestX < best) {
                bestX = best;
                bestDt = dt;
            }
        }
        LOG(INFO) << "2 bestT=" << bestX << " bestDT=" << bestDt;
#endif
        *dto = bestDt;

        bestX = 0;
        double bestDv0x = 0;
        for (int i = 0; i <= cnt; i++) {
            double ndv0x = dv0x + (i / (double)cnt - 0.5) * scale;
            willSucceed2(sim, oxa, oya,
                         start, bestDt, ndv0x, dv0y, &best);
            if (bestX <= best) {
                bestX = best;
                bestDv0x = ndv0x;
            }
        }
        LOG(INFO) << "2 bestX=" << bestX << " bestDv0x=" << bestDv0x;

        double bestY = 0;
        double bestDv0y = 0;
        for (int i = 0; i <= cnt; i++) {
            double ndv0y = dv0y + (i / (double)cnt - 0.5) * scale;
            willSucceed2(sim, oxa, oya,
                         start, bestDt, dv0x, ndv0y, &best);
            if (bestY <= best) {
                bestY = best;
                bestDv0y = ndv0y;
            }
        }
        LOG(INFO) << "2 bestY=" << bestX << " bestDv0y=" << bestDv0y;

        willSucceed2(sim, oxa, oya,
                     start, bestDt, bestDv0x, bestDv0y, &best);

        LOG(INFO) << "2 best=" << best << " bestX="
                  << bestX << " bestY=" << bestY;
        if (best < bestX && bestY <= bestX) {
            bestDv0y = dv0y;
            best = bestX;
            LOG(INFO) << "X is best " << best;
        }
        else if (best < bestY && bestX <= bestY) {
            bestDv0x = dv0x;
            best = bestY;
            LOG(INFO) << "Y is best " << best;
        }
        *bestDv0xo = bestDv0x;
        *bestDv0yo = bestDv0y;

        return best;
    }

    bool decidePlan(const vector<double>& xa, const vector<double>& ya,
                    const vector<double>& oxa, const vector<double>& oya,
                    const vector<double>& ovxa, const vector<double>& ovya,
                    int start, double t,
                    double dv0x, double dv0y, double dvtx, double dvty) {
        double px, py;
        if (start > 0) {
            px = xa[start-1];
            py = ya[start-1];
        }
        else {
            px = this->px;
            py = this->py;
        }

        double best;
        if (willSucceed(px, py, xa[start], ya[start], oxa, oya,
                        start, t, dv0x, dv0y, dvtx, dvty, &best)) {
            return false;
        }
        LOG(INFO) << "best=" << best << " dv0x=" << dv0x << " dv0y=" << dv0y;

        double bestDv0x, bestDv0y;
        best = adjustDv(xa, ya, oxa, oya, px, py, start, t, dv0x, dv0y,
                        10, 10, &bestDv0x, &bestDv0y);
        best = adjustDv(xa, ya, oxa, oya, px, py, start, t, bestDv0x, bestDv0y,
                        10, 5, &bestDv0x, &bestDv0y);
        best = adjustDv(xa, ya, oxa, oya, px, py, start, t, bestDv0x, bestDv0y,
                        10, 1, &bestDv0x, &bestDv0y);
        best = adjustDv(xa, ya, oxa, oya, px, py, start, t, bestDv0x, bestDv0y,
                        10, 0.5, &bestDv0x, &bestDv0y);
        best = adjustDv(xa, ya, oxa, oya, px, py, start, t, bestDv0x, bestDv0y,
                        10, 0.2, &bestDv0x, &bestDv0y);
        best = adjustDv(xa, ya, oxa, oya, px, py, start, t, bestDv0x, bestDv0y,
                        100, 0.1, &bestDv0x, &bestDv0y);

        if (best >= 1000) {
            nextHohmann = turn + start + 1000;
            LOG(INFO) << "Our plan was not good. Wait for the next chance: "
                      << nextHohmann;
            return true;
        }

        int firstBoost = turn+start;

        Simulator sim(px, py, xa[start], ya[start]);
        sim.step(bestDv0x, bestDv0y);
        int i;
        for (i = 1;; i++) {
            CHECK_LT(start+i, oxa.size());

            double d = dist(sim.x - oxa[start+i], sim.y - oya[start+i]);
            if (d == best) break;

            sim.save();
            sim.step();
        }

        sim.restore();

        double dvx = -(ovxa[start+i] - sim.vx);
        double dvy = -(ovya[start+i] - sim.vy);
        double dt = 0;
        LOG(INFO) << "Changing speed adjustment: " << POS(dvx, dvy);
        best = adjustDv2(sim, xa, ya, oxa, oya, px, py, start+i, t, dvx, dvy,
                         20, 1, &dvx, &dvy, &dt);
        best = adjustDv2(sim, xa, ya, oxa, oya, px, py, start+i, t, dvx, dvy,
                         20, 0.1, &dvx, &dvy, &dt);
        best = adjustDv2(sim, xa, ya, oxa, oya, px, py, start+i, t, dvx, dvy,
                         20, 0.01, &dvx, &dvy, &dt);
        best = adjustDv2(sim, xa, ya, oxa, oya, px, py, start+i, t, dvx, dvy,
                         20, 0.001, &dvx, &dvy, &dt);
        best = adjustDv2(sim, xa, ya, oxa, oya, px, py, start+i, t, dvx, dvy,
                         20, 0.0001, &dvx, &dvy, &dt);
        best = adjustDv2(sim, xa, ya, oxa, oya, px, py, start+i, t, dvx, dvy,
                         20, 0.00001, &dvx, &dvy, &dt);
        LOG(INFO) << "Use dvx=" << dvx << ", dvy=" << dvy << ", dt=" << dt;

        vector<double> planX;
        vector<double> planY;
        planX.push_back(dvx * (1 - dt));
        planY.push_back(dvy * (1 - dt));
        planX.push_back(dvx * dt);
        planY.push_back(dvy * dt);

        sim.restore();
        sim.step(dvx * (1 - dt), dvy * (1 - dt));
        sim.step(dvx * dt, dvy * dt);

        for (int j = 0; j < 900; j++) {
            sim.save();
            if (best > 1500) break;
            best = adjustDv2(sim, xa, ya, oxa, oya, px, py, start+i+2+j, t, 0, 0,
                             20, 0.01, &dvx, &dvy, &dt);
            best += j;
            planX.push_back(dvx);
            planY.push_back(dvy);
            sim.step(dvx, dvy);
        }

        if (best < 1000) {
            nextHohmann = turn + start + 1000;
            LOG(INFO) << "Our plan was not good2. Wait for the next chance: "
                      << nextHohmann;
            return true;
        }

        addPlan(firstBoost, 2, bestDv0x);
        addPlan(firstBoost, 3, bestDv0y);
        addPlan(firstBoost+1, 2, 0);
        addPlan(firstBoost+1, 3, 0);

        CHECK_EQ(planX.size(), planY.size());
        for (int j = 0; j < planX.size(); j++) {
            addPlan(turn+start+i-1+j, 2, planX[j]);
            addPlan(turn+start+i-1+j, 3, planY[j]);
        }
/*
        addPlan(turn+start+i-1, 2, dvx * (1 - dt));
        addPlan(turn+start+i-1, 3, dvy * (1 - dt));
        addPlan(turn+start+i, 2, dvx * dt);
        addPlan(turn+start+i, 3, dvy * dt);
*/

/*
        for (int j = 0; j < 5000; j++) {
            i++;
            sim.step(-dvx, -dvy);
            dvx = ovxa[start+i] - sim.vx;
            dvy = ovya[start+i] - sim.vy;
            LOG(INFO) << POS(dvx, dvy);
            addPlan(turn+start+i-1, 2, -dvx);
            addPlan(turn+start+i-1, 3, -dvy);
        }
*/

        addPlan(turn+start+i+planX.size()-1, 2, 0);
        addPlan(turn+start+i+planY.size()-1, 3, 0);

        //nextHohmann = turn+start+i+5000;

        arrival = turn+start+i;
        LOG(INFO) << "We start at " << (int)(turn+start)
                  << " with " << POS(dv0x, dv0y)
                  << " and will arrive at " << (int)(arrival)
                  << " with " << POS(dvtx, dvty)
                  << " target should be in " << POS(oxa[start+i], oya[start+i])
                  << " next trip will be " << nextHohmann;

        return true;
    }

    double px, py;
    double pox, poy;
    int nextHohmann;
    bool clockwise;
    int arrival;
};

Actuator* createActuator3() {
    return new Actuator3();
}
