#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <dlfcn.h>

#include <vector>

#include <gflags/gflags.h>

#include "vm.h"

DEFINE_bool(dump_memory, false, "");

using namespace std;

typedef unsigned int uint;

enum {
    ADD, SUB, MUL, DIV, OUT, PHI, NOP, CMPZ, SQRT, COPY, INPUT,
};
#if 0
static const char* OP_STR[] = {
    "ADD", "SUB", "MUL", "DIV", "OUT", "PHI",
    "NOP", "CMPZ", "SQRT", "COPY", "INPUT",
};
#endif

struct Op {
    int op;
    int r;
    union {
        int i;
        int r2;
    };
};

struct In {
    uint addr;
    double value;
};

extern "C" {
    typedef void setInput_func_t(int addr, double value);
    typedef double* getOutput_func_t();
    typedef void setMemory_func_t(double* m);
    typedef void step_func_t();
}

class VMSO : public VM {
public:
    VMSO(const char* file) {
        FILE* fp = fopen(file, "rb");
        fseek(fp, 0, SEEK_END);
        long len = ftell(fp);
        fseek(fp, 0, SEEK_SET);
        uint* buf = (uint*)malloc(len);
        fread(buf, 1, len, fp);
        fclose(fp);

        Op* ops = (Op*)malloc(len);
        len /= 4;
        for (int i = 0; i < len; i += 3) {
            uint inst = buf[i % 2 == 0 ? i + 2 : i];
            uint op = inst >> 28;
            uint r1, r2;
            if (op) {
                op = op - 1;
                //DCHECK_LT(op, NOP);
                r1 = (inst >> 14) & 0x3fff;
                r2 = inst & 0x3fff;
            }
            else {
                op = NOP + (inst >> 24);
                //DCHECK_LE(op, INPUT);
                r1 = inst & 0x3fff;
                r2 = (inst >> 21) & 7;
            }
            int a = i / 3;
            ops[a].op = op;
            ops[a].r = r1;
            ops[a].r2 = r2;
        }

        for (int i = 0; i < len; i += 3) {
            uint* p = buf + (i % 2 == 0 ? i : i + 1);
            mem[i / 3] = *(double*)p;
        }

        if (FLAGS_dump_memory) {
            for (int i = 0; i < len; i += 3) {
                printf("%d %lf %g\n", i / 3, mem[i / 3], mem[i / 3]);
            }
            exit(0);
        }

        free(buf);

        compile(ops, len / 3);
        setMemory_func(mem);
    }

    void compile(Op* ops, long len) {
        char tmp[256] = "tmp/vm_XXXXXX";
        mkstemp(tmp);
        char buf[256];
        sprintf(buf, "%s.c", tmp);
        FILE* fp = fopen(buf, "wb");
        fprintf(fp,
                "#include <math.h>\n"
                "static double* mem = 0;\n"
                "static double in[16384];\n"
                "static double out[16384];\n"
                "static int status = 0;\n"
                "void setInput(int addr, double value) { in[addr] = value; }\n"
                "double* getOutput() { return out; }\n"
                "void setMemory(double* m) { mem = m; }\n"
                "void step() {\n"
            );
        for (int i = 0; i < len; i++) {
            switch (ops[i].op) {
            case ADD:
                fprintf(fp,
                        "  mem[%d] = mem[%d] + mem[%d];\n",
                        i, ops[i].r, ops[i].r2
                    );
                break;
            case SUB:
                fprintf(fp,
                        "  mem[%d] = mem[%d] - mem[%d];\n",
                        i, ops[i].r, ops[i].r2
                    );
                break;
            case MUL:
                fprintf(fp,
                        "  mem[%d] = mem[%d] * mem[%d];\n",
                        i, ops[i].r, ops[i].r2
                    );
                break;
            case DIV:
                fprintf(fp,
                        "{  double d = mem[%d];\n"
                        "  mem[%d] = d != 0 ? mem[%d] / d : 0.0; }\n",
                        ops[i].r2, i, ops[i].r
                    );
                break;
            case OUT:
                fprintf(fp,
                        "  out[%d] = mem[%d];\n",
                        ops[i].r, ops[i].r2
                    );
                break;
            case PHI:
                fprintf(fp,
                        "  mem[%d] = status ? mem[%d] : mem[%d];\n",
                        i, ops[i].r, ops[i].r2
                    );
                break;
            case NOP:
                break;
            case CMPZ:
                switch (ops[i].i) {
                case 0: /* LTZ */
                    fprintf(fp,
                            "  status = mem[%d] < 0.0;\n",
                            ops[i].r
                        );
                    break;
                case 1: /* LEZ */
                    fprintf(fp,
                            "  status = mem[%d] <= 0.0;\n",
                            ops[i].r
                        );
                    break;
                case 2: /* EQZ */
                    fprintf(fp,
                            "  status = mem[%d] == 0.0;\n",
                            ops[i].r
                        );
                    break;
                case 3: /* GEZ */
                    fprintf(fp,
                            "  status = mem[%d] >= -0.0;\n",
                            ops[i].r
                        );
                    break;
                case 4: /* GTZ */
                    fprintf(fp,
                            "  status = mem[%d] > -0.0;\n",
                            ops[i].r
                        );
                    break;
                default:
                    assert(0);
                }
                break;
            case SQRT:
                fprintf(fp,
                        "  mem[%d] = sqrt(mem[%d]);\n",
                        i, ops[i].r
                    );
                break;
            case COPY:
                fprintf(fp,
                        "  mem[%d] = mem[%d];\n",
                        i, ops[i].r
                    );
                break;
            case INPUT:
                fprintf(fp,
                        "  mem[%d] = in[%d];\n",
                        i, ops[i].r
                    );
                break;
            default:
                assert(0);

            }
        }
        fprintf(fp,
                "}\n"
            );
        fclose(fp);

        sprintf(buf, "gcc -fPIC -O -g -shared %s.c -o %s.so", tmp, tmp);
        system(buf);

        sprintf(buf, "%s.so", tmp);
        handle = dlopen(buf, RTLD_NOW);
        setInput_func = (setInput_func_t*)dlsym(handle, "setInput");
        getOutput_func = (getOutput_func_t*)dlsym(handle, "getOutput");
        setMemory_func = (setMemory_func_t*)dlsym(handle, "setMemory");
        step_func = (step_func_t*)dlsym(handle, "step");
    }

    virtual ~VMSO() {
        dlclose(handle);
    }

    virtual void setInput(int addr, double value) {
        setInput_func(addr, value);
    }

    virtual double* getOutput() {
        return getOutput_func();
    }

    virtual void step() {
        step_func();
    }

    void* handle;
    double mem[16384];
    setInput_func_t* setInput_func;
    getOutput_func_t* getOutput_func;
    setMemory_func_t* setMemory_func;
    step_func_t* step_func;
};

VM* createVM(const char* file) {
    return new VMSO(file);
}

